# -*- coding: UTF-8 -*-
__version__ = "1.0"

from netapp_element import NetAppObject
import base64
import httplib
import xml.parsers.expat
from netapp_constants import NetAppConstants
from netapp_exceptions import FailedParamException
from netapp_exceptions import InvokeException
from netapp_exceptions import ParseException
import socket
import sys
socket_ssl_attr = True
if not hasattr(socket, 'ssl'):
    socket_ssl_attr = False
ssl_import = False
try:
    import ssl
    ssl_import = True
except ImportError:
    pass
CNETR_VERSION = "9.3"
CNETR_LANGUAGE = "Python"
nmsdk_app_name = ""


class NetAppConnector(object):
    """
    netapp device connector:
    :only support for filer device using login-password verification
    """
    def __init__(self, server, major_version, minor_version):
        """
        initiator
        """

        self.server = server
        self.major_version = major_version
        self.minor_version = minor_version
        self.url = NetAppConstants.NETAPP_SERVER_URLS[NetAppConstants.NETAPP_SERVER_TYPE_FILER]
        self.dtd = NetAppConstants.NETAPP_SERVER_DTDS[NetAppConstants.NETAPP_SERVER_TYPE_FILER]
        self.server_type = NetAppConstants.NETAPP_SERVER_TYPE_FILER
        self.login_type = NetAppConstants.NETAPP_SIGN_IN_METHOD_LOGIN
        self.transport_type = NetAppConstants.CONNECTION_TYPE_HTTPS
        self.port = 443
        # default timeout (seconds)
        self.__timeout = 3 * 60
        self.__user, self.__password = "", ""
        self.vfiler = self.open_debug = self.output_in_xml = self.originator_id = ""
        self.cert_file = self.key_file = self.ca_file = None
        self.need_server_auth = self.need_cn_verification = self.need_cba = False
        self.ZAPI_stack = []
        self.ZAPI_atts = {}

    def set_login_type(self, login_type):
        """
        :param login_type: style only supports hosts\logon\certificate
        :return:
        """

        if login_type not in NetAppConstants.NETAPP_SIGN_IN_METHODS:
            raise FailedParamException("failed to parse login types, unrecognized type: " + login_type)

        if login_type == NetAppConstants.NETAPP_SIGN_IN_METHOD_CERTIFICATE:
            if not ssl_import:
                raise FailedParamException("cannot use 'ssl' module as is not imported.")
            if not socket_ssl_attr:
                raise FailedParamException("cannot use 'socket' module as is not compiled with SSL support.")
            self.set_transport_type(NetAppConstants.CONNECTION_TYPE_HTTPS)
            self.need_cba = True
            self.set_server_cert_verification(True)
        else:
            self.need_cba = False
            self.set_server_cert_verification(False)
        self.login_type = login_type
        return None

    def get_login_type(self):
        """
        Get the log on method
        """
        return self.login_type

    def set_login_params(self, user, password):
        """
        :param user: user names
        :param password: current connected device password
        :return:  no return
        """
        self.__user = user
        self.__password = password

    def set_server_type(self, server_type):
        """
        set  remote server type: only  'FILER' or 'DFM' or 'OCUM' is  allowed
        """
        server_type = server_type.upper()
        if server_type not in NetAppConstants.NETAPP_SERVER_TYPES:
            raise FailedParamException("wrong server type detected: " + server_type)
        self.url = NetAppConstants.NETAPP_SERVER_URLS[server_type]
        self.dtd = NetAppConstants.NETAPP_SERVER_DTDS[server_type]
        if NetAppConstants.NETAPP_SERVER_TYPE_OCUM == server_type:
            self.transport_type = NetAppConstants.CONNECTION_TYPE_HTTPS
        self.port = NetAppConstants.NETAPP_SERVER_DEFAULT_PORT[server_type]
        if NetAppConstants.NETAPP_SERVER_TYPE_DFM == server_type and \
                NetAppConstants.CONNECTION_TYPE_HTTPS == self.transport_type:
            self.port = 8488
        self.server_type = server_type
        return

    def get_server_type(self):
        """
        Get the type of server this server connection applies to.
        """
        return self.server_type

    def set_port(self, port):
        """
        set the port number manually
        """
        self.port = port

    def get_port(self):
        """
        Retrieve the port used for the remote server.
        """
        return self.port

    def set_timeout(self, timeout):
        """
        set timeout of http connection 
        """
        if str(timeout).isdigit():
            self.__timeout = timeout

    def set_vserver(self, vserver):
        """
        For vserver-tunneling,this function internally sets the vfiler name.
        :param vserver: the vserver name
        :return:  on   return
        """
        if self.major_version >= 1 and self.minor_version >= 15:
            self.vfiler = vserver
            return
        else:
            raise FailedParamException("ONTAPI version must be at least 1.15 to send API to a vserver")

    def set_originator_id(self, originator_id):
        """
        Function to set the originator_id before executing any ONTAP API.
        """
        self.originator_id = originator_id
        return

    def get_originator_id(self):
        """
        Gets the originator_id for the given server context on which the
        ONTAP API commands get invoked.
        """
        return self.originator_id

    def set_transport_type(self, trans_type):
        """
        Override the default transport type.  The valid transport
        type are currently 'HTTP' and 'HTTPS'.
        """
        trans_type = trans_type.upper()
        if trans_type not in NetAppConstants.COMMON_CONNECTION_PROTOCALS:
            raise FailedParamException("unknown transport Type Exception,only http & https are allowed")
        self.transport_type = trans_type
        if NetAppConstants.CONNECTION_TYPE_HTTP == trans_type:
            if self.server_type == NetAppConstants.NETAPP_SERVER_TYPE_OCUM:
                raise FailedParamException("Server '" + self.server_type + "' does not support '" + trans_type)
            if NetAppConstants.NETAPP_SERVER_TYPE_DFM == self.server_type:
                self.port = NetAppConstants.NETAPP_SERVER_DEFAULT_PORT[NetAppConstants.NETAPP_SERVER_TYPE_DFM]
            else:
                self.port = NetAppConstants.NETAPP_SERVER_DEFAULT_PORT[NetAppConstants.NETAPP_SERVER_TYPE_NETCACHE]
        else:
            if not socket_ssl_attr:
                raise FailedParamException(trans_type + " cannot be used as SSL not in  'socket' module.")
            if NetAppConstants.NETAPP_SERVER_TYPE_DFM == self.server_type:
                self.port = 8488
            else:
                self.port = NetAppConstants.NETAPP_SERVER_DEFAULT_PORT[NetAppConstants.NETAPP_SERVER_TYPE_FILER]
        return

    def get_transport_type(self):
        """
        Retrieve the transport used for this connection.
        """
        return self.transport_type

    def set_debug_enabled(self):
        """
        Set the style of debug.
        """
        self.open_debug = NetAppConstants.NETAPP_DEBUG_STYLE
        return

    def is_debugging(self):
        """
        :check whether the debug switch is on.
        """
        if self.open_debug:
            return True
        else:
            return False

    def get_raw_xml_output(self):
        """
        Return the raw XML output.
        """
        return self.output_in_xml

    def set_raw_xml_output(self, xml_output):
        """
        Save the raw XML output.
       """
        self.output_in_xml = xml_output

    def is_use_https(self):
        """
        Determines whether https is enabled.
       """
        if NetAppConstants.CONNECTION_TYPE_HTTPS == self.transport_type:
            return True
        else:
            return False

    def call_remote(self, req):
        """
        Submit an XML request encapsulated as
        an NetAppObject and return as one too.
        """
        originator_id = self.originator_id
        xmlrequest = req.encapsulate_cmd()
        vfiler_req = ""
        originator_id_req = ""
        nmsdk_app_req = ""
        try:
            httplib.socket.setdefaulttimeout(self.__timeout)
            '''
            seal http header:
            |---------------------------
            |http title
            |---------------------------
            |http header
            |---------------------------
            |http body
            '''
            if self.transport_type == NetAppConstants.CONNECTION_TYPE_HTTP:
                connection = httplib.HTTPConnection(self.server, port = self.port)
            else:
                if self.need_cba or self.need_server_auth:
                    raise FailedParamException("certificate based authentication is not currently supported  ")
                else:
                    connection = httplib.HTTPSConnection(self.server, port = self.port)

            connection.putrequest('POST', self.url)
            connection.putheader('Content-type', 'text/xml; charset="UTF-8"')

            if self.get_login_type() != NetAppConstants.NETAPP_SIGN_IN_METHOD_HOSTS:
                base64string = base64.encodestring('%s:%s' % (self.__user, self.__password))[:-1]
                auth_header = "Basic %s" % base64string
                connection.putheader("Authorization", auth_header)
            if self.vfiler:
                vfiler_req = ' vfiler="' + self.vfiler + '"'

            if originator_id:
                originator_id_req = ' originator_id="' + originator_id + '"'

            if nmsdk_app_name:
                nmsdk_app_req = ' nmsdk_app="' + nmsdk_app_name + '"'

            content = '<?xml version="1.0" encoding="utf-8"?>\n' + \
                      '<!DOCTYPE netapp SYSTEM "' + self.dtd + '">' + \
                      '<netapp' + vfiler_req + originator_id_req + ' version="' + str(self.major_version) + \
                      '.' + str(self.minor_version) + '"' + ' xmlns="' + 'http://www.netapp.com/filer/admin' + '"' \
                      + ' nmsdk_version="' + CNETR_VERSION + '"' + ' nmsdk_platform="' + self.get_platform_info() + \
                      '"' + ' nmsdk_language="' + CNETR_LANGUAGE + "\"" + nmsdk_app_req + ">" \
                      + xmlrequest + '</netapp>'
            connection.putheader("Content-length", len(content))
            connection.endheaders()
            connection.send(content)
        except socket.error:
            message = sys.exc_info()
            raise InvokeException("exception occurred sending http request on  invoke:" + str(message[1]))

        response = connection.getresponse()

        if not response:
            connection.close()
            raise InvokeException("No response received")

        if response.status == NetAppConstants.HTTP_CONN_CODE_UNAUTHORIZED:
            connection.close()
            raise InvokeException("Authorization failed")

        xml_response = response.read()

        connection.close()
        return self.__parse_2_netapp_obj(xml_response)

    def request(self, api, *arg):
        """
        the api's arguments must not be odd! it shall be  in  key-value format
        """

        num_parms = len(arg)

        if num_parms & 1:
            raise InvokeException("parameters must not be odd!")

        send_obj = NetAppObject(api)
        poi = 0

        while poi < num_parms:
            key = arg[poi]
            poi = poi + 1
            value = arg[poi]
            poi = poi + 1
            send_obj.add_element(NetAppObject(key, value))

        return self.call_remote(send_obj)

    def set_vfiler(self, vfiler_name):
        """
        Sets the vfiler name. This function is used
        for vfiler-tunneling.
        """
        if self.major_version >= 1 and self.minor_version >= 7:
            self.vfiler = vfiler_name
            return True
        return False

    def set_client_cert_and_key(self, cert_file, key_file):
        """
        Sets the client certificate and key files that are required for client authentication
        by the server using certificates. If key file is not defined, then the certificate file 
        will be used as the key file.
        """
        self.cert_file = cert_file
        if key_file:
            self.key_file = key_file
        else:
            self.key_file = cert_file

    def set_ca_certs(self, ca_file):
        """
        Specifies the certificates of the Certificate Authorities that are
        trusted by this application and that will be used to verify the server certificate.
        """
        self.ca_file = ca_file

    def set_server_cert_verification(self, enable):
        """
        Enables or disables server certificate verification.
        """
        if enable not in [True, False]:
            raise FailedParamException("set_server_cert_verification argument must be True or False")
        if not self.is_use_https():
            return
        if enable and not ssl_import:
            raise FailedParamException("certificate verification cannot be used as 'ssl' module is not imported.")
        self.need_server_auth = enable
        self.need_cn_verification = enable
        return None

    def is_server_cert_verification_enabled(self):
        """
        Determines whether server certificate verification is enabled or not.
        """
        return self.need_server_auth

    def set_hostname_verification(self, enable):
        """
        Enables or disables hostname verification during server certificate verification.
        """
        if enable not in [True, False]:
            raise FailedParamException("set_hostname_verification argument must be True or False")
        if not self.need_server_auth:
            raise FailedParamException("server certificate verification is not enabled")
        self.need_cn_verification = enable
        return

    def is_hostname_verification_enabled(self):
        """
        Determines whether hostname verification is enabled or not.
        Returns True if it is enabled, else returns False
        """
        return self.need_cn_verification

    def __start_element(self, name, attrs):
        """
        encapsulate request element
        """
        element = NetAppObject(name)
        self.ZAPI_stack.append(element)
        self.ZAPI_atts = {}
        attr_name = list(attrs.keys())
        attr_value = list(attrs.values())
        poi = 0
        for att in attr_name:
            val = attr_value[poi]
            poi += 1
            self.ZAPI_atts[att] = val
            element.put_element_attr(att, val)

    def __end_element(self, name):
        """
        encapsulate request element
        """
        stack_len = len(self.ZAPI_stack)

        if stack_len > 1:
            n = self.ZAPI_stack.pop(stack_len - 1)
            i = len(self.ZAPI_stack)
            if i != stack_len - 1:
                #"pop did not work!!!!\n")
                pass
            self.ZAPI_stack[i - 1].add_element(n)

    def __add_element_content(self, data):
        """
        This is a private function, not to be called from outside NaElement
        """
        zapi_len = len(self.ZAPI_stack)
        data = NetAppObject.escape_html(data)
        self.ZAPI_stack[zapi_len - 1].add_content(data)

    def __parse_2_netapp_obj(self, xml_response):
        """

        """
        parser_obj = xml.parsers.expat.ParserCreate()
        parser_obj.StartElementHandler = self.__start_element
        parser_obj.EndElementHandler = self.__end_element
        parser_obj.CharacterDataHandler = self.__add_element_content
        parser_obj.Parse(xml_response, 1)
        stack_len = len(self.ZAPI_stack)

        if stack_len <= 0:
            raise ParseException("parse_xml-no elements on stack")

        node = self.ZAPI_stack.pop(stack_len - 1)
        if node.element['name'] != NetAppConstants.NETAPP_ELEMENT_NETAPP:
            raise ParseException("parse_xml - Expected <netapp> element but got " + node.element['name'])
        results = node.get_element_recursively("results")
        if results is None:
            raise ParseException("parse_xml - No results element in output!")
        return results

    def __parse_raw_xml(self, xmlrequest):
        """
        This is a private function, not to be called from outside NaElement
        """
        parser_obj = xml.parsers.expat.ParserCreate()
        parser_obj.StartElementHandler = self.__start_element
        parser_obj.EndElementHandler = self.__end_element
        parser_obj.CharacterDataHandler = self.__add_element_content
        parser_obj.Parse(xmlrequest, 1)
        stack_len = len(self.ZAPI_stack)

        if stack_len <= 0:
            raise ParseException("parse_xml but found no elements on stack")
        one_node = self.ZAPI_stack.pop(stack_len - 1)
        return one_node

    @staticmethod
    def set_application_name(app_name):
        """
        Sets the name of the client application.
        """
        global nmsdk_app_name
        nmsdk_app_name = app_name

    @staticmethod
    def get_application_name():
        """
        Returns the name of the client application.
        """
        global nmsdk_app_name
        return nmsdk_app_name

    @staticmethod
    def get_platform_info():
        """
        Returns the platform information.
        """
        system_type = "Unknown"
        os_info = ""

        try:
            import platform
            system_type = platform.system()
            if system_type == "Windows" or system_type == "Microsoft":
                system_type = "Windows"
                import _winreg
                handle = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion")
                os_name, reg_value = _winreg.QueryValueEx(handle, "ProductName")
                _winreg.CloseKey(handle)
                handle = _winreg.OpenKey(
                    _winreg.HKEY_LOCAL_MACHINE, "SYSTEM\\ControlSet001\\Control\\Session Manager\\Environment")
                processor, reg_value = _winreg.QueryValueEx(handle, "PROCESSOR_ARCHITECTURE")
                _winreg.CloseKey(handle)
                os_info = os_name + " " + processor
        except:
            os_info = system_type
        return os_info
