# -*- coding: UTF-8 -*-
import re
from netapp_exceptions import ParseException
from netapp_exceptions import ElementException
from netapp_constants import NetAppConstants


class NetAppObject(object):
    """
    to invoke ONTAPI API calls, a XML pretty string must be encapsulated
    """

    def __init__(self, name, value=None):
        """
        Construct a new NetAppObject.  The 'value' parameter is
        optional for top level elements.
        """

        self.element = {'name': name, 'content': "", 'children': [], 'attrkeys': [], 'attrvals': []}
        if value is not None:
            self.element['content'] = value

    def get_results_status(self):
        """
        Indicates success or failure of API call.
        Returns either 'passed' or 'failed'.
        """
        result_status = self.get_element_attr("status")
        if result_status == "passed":
            return "passed"

        return "failed"

    def get_failure_reason(self):
        """
        :return the description of a failure.
        """
        result_status = self.get_element_attr("status")
        if result_status == "passed":
            return None
        result_status = self.get_element_attr("reason")
        if not result_status:
            return "No reason given"
        return str(result_status)

    def get_element_recursively(self, name):
        """
        get sub-element of current node, recursively search all elements
        under current element
        """

        xml_node = self.element['children']
        for cur_node in xml_node:
            if name == cur_node.element['name']:
                return cur_node
        return None

    def set_content(self, content):
        """
        Set the element's value to 'content'.
        """
        self.element['content'] = content

    def add_content(self, content):
        """
        """
        self.element['content'] += content
        return

    def add_element(self, element):
        """
        :param element sub-element of NetAppObject type
        """
        elements = self.element['children']
        elements.append(element)
        self.element['children'] = elements
    def put_element_attr(self, key, value):
        """
        add new  attribute to current element
        """

        arr = self.element['attrkeys']
        arr.append(key)
        self.element['attrkeys'] = arr
        arr = self.element['attrvals']
        arr.append(value)
        self.element['attrvals'] = arr

    def get_element_attr(self, key):
        """
        get element's attribute in k-v format
        """

        keys = self.element['attrkeys']
        vals = self.element['attrvals']
        j = 0

        for i in keys:
            if i == key:
                return vals[j]
            j = j + 1
        return None

    def get_children(self):
        """
        :returns all children under current element
        """
        children_element = self.element['children']
        return children_element

    def to_pretty_xml(self, indent=""):
        """
        :summary convert to xml-pretty style string
        """
        name = self.element['name']
        result = indent + '<' + name
        keys = self.element['attrkeys']
        vals = self.element['attrvals']
        for poi in range(len(keys)):
            result += " " + str(keys[poi]) + '="' + str(vals[poi]) + '"'
        result += '>'
        children = self.element['children']

        if len(children) > 0:
            result += '\n'
            for child in children:
                if not re.search('NetAppObject', str(child.__class__), re.I):
                    raise ParseException("Unexpected reference found, expected NetAppObject")
                result += child.to_pretty_xml(indent + "\t")

        result += str(self.element['content'])

        if len(children) > 0:
            result += indent
        result += "</" + name + ">\n"
        return result

    def encapsulate_cmd(self):
        """
        encapsulate ontap api format, convert into xml pretty format
        """
        name_tag = str(self.element['name'])
        result = "<" + name_tag
        keys = self.element['attrkeys']
        vals = self.element['attrvals']
        for poi in range(len(keys)):
            result += " " + str(keys[poi]) + '="' + str(vals[poi]) + '"'
        result += '>'
        children = self.element['children']

        for child in children:
            if not re.search("NetAppObject", str(child.__class__), re.I):
                raise ElementException("Unexpected reference to encoded String, expected NetAppObject.")

            result += child.encapsulate_cmd()

        content = str(self.element['content'])
        result += content + "</" + name_tag + ">"
        return result

# ------------------------------------------------------------#
#
# routines beyond this point are "private"
#
# ------------------------------------------------------------#
    @staticmethod
    def escape_html(cont):
        """
        This is a private function, not to be called externally.
        This method converts reserved HTML characters to corresponding entity names.
        """
        cont = re.sub(r'&', '&amp;', cont, count=0)
        cont = re.sub(r'<', '&lt;', cont, count=0)
        cont = re.sub(r'>', '&gt;', cont, count=0)
        cont = re.sub(r"'", '&apos;', cont, count=0)
        cont = re.sub(r'"', '&quot;', cont, count=0)

        """ 
        The existence of '&' (ampersand) sign in entity names implies that multiple calls
        to this function will result in non-idempotent encoding. So, to handle such situation
        or when the input itself contains entity names, we reconvert such recurrences to
        appropriate characters.
        """
        cont = re.sub(r'&amp;amp;', '&amp;', cont, count=0)
        cont = re.sub(r'&amp;lt;', '&lt;', cont, count=0)
        cont = re.sub(r'&amp;gt;', '&gt;', cont, count=0)
        cont = re.sub(r'&amp;apos;', '&apos;', cont, count=0)
        cont = re.sub(r'&amp;quot;', '&quot;', cont, count=0)
        return cont

